########################################################################
# BigFloat for Tcl
# Copyright (C) 2003-2004  ARNOLD Stephane
#
# BIGFLOAT LICENSE TERMS
#
# This software is copyrighted by Stephane ARNOLD, (stephanearnold <at> yahoo.fr).
# The following terms apply to all files associated
# with the software unless explicitly disclaimed in individual files.
#
# The authors hereby grant permission to use, copy, modify, distribute,
# and license this software and its documentation for any purpose, provided
# that existing copyright notices are retained in all copies and that this
# notice is included verbatim in any distributions. No written agreement,
# license, or royalty fee is required for any of the authorized uses.
# Modifications to this software may be copyrighted by their authors
# and need not follow the licensing terms described here, provided that
# the new terms are clearly indicated on the first page of each file where
# they apply.
#
# IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
# DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
# IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.
#
# GOVERNMENT USE: If you are acquiring this software on behalf of the
# U.S. government, the Government shall have only "Restricted Rights"
# in the software and related documentation as defined in the Federal
# Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
# are acquiring the software on behalf of the Department of Defense, the
# software shall be classified as "Commercial Computer Software" and the
# Government shall have only "Restricted Rights" as defined in Clause
# 252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
# authors grant the U.S. Government and others acting in its behalf
# permission to use and distribute the software in accordance with the
# terms specified in this license.
#
########################################################################

# this line helps when I want to source this file again and again
catch {namespace delete ::math::bigfloat}

# private namespace
# this software works only with Tcl v8.4 and higher
# because it uses 64-bit integers.
namespace eval ::math::bigfloat {
    # some constants
    variable karatsubaLimit
    set karatsubaLimit 13
    # max length of an integer in the internal representation of a BigInt
    variable segmentSize
    set segmentSize 9
    variable segmentZero
    set segmentZero 1000000000
    # cached constants
    # ln(2) with arbitrary precision
    variable log2
    # ln(10) with arb. precision
    variable log10
    # Pi with arb. precision
    variable _pi0
}

################################################################################
# PART I
# PROCEDURES DEALING WITH BIG SIGNED INTEGERS
#
# When a signed integer is used, the sign is the first element of the list, then
# come the chunks of the entire integer. Each chunk is 8 digits long.
# Chunks are [lappend]ed in reverse order. For example : -100000002
# is decomposed into a sign (-1) and two chunks (1 and 00000002, or 1 and 2)
# is represented internally as : [list -1 2 1] (heading zeros are trimmed)
# By convetion, when the parameter "raw" is set to 1, unsigned integers are used.
# They do not have the first element as the sign.
# For example : 100000002 is just [list 2 1]
################################################################################


################################################################################
# global utility function
# trims all zeroes in the internal representation of a BigInt
# it "normalizes" the BigInt
################################################################################
proc ::math::bigfloat::Dec {number {raw 0}} {
    # fix the signed BigInt problem
    if {!$raw} {return [concat [lindex $number 0] [Dec [lrange $number 1 end] 1]]}
    set l [llength $number]
    # trivial case when number is a void list
    if {$l==0} {
        return {0}
    }
    set result {}
    # copy number in result...
    foreach i $number {
        # ... while trimming zeros
        lappend result [trimleft $i]
    }
    # delete zeros that are at the head of the number (at the end of
    # the internal list representation)
    while {[lindex $result end]==0} {
        set result [lrange $result 0 end-1]
    }
    if {[llength $result]==0} {lappend result 0}
    return $result
}


################################################################################
# computes the number of digits in a BigInt
################################################################################
proc ::math::bigfloat::length {x {raw 0}} {
    set x [Dec $x $raw]
    if {!$raw} {
        return [expr {([llength $x]-2)*$::math::bigfloat::segmentSize+[string length\
                    [trimleft [lindex $x end]]]}]
        
    }
    # llength $x = chunk number of $segmentSize digits
    # the last integer may have less than $segmentSize digits
    # (and the last in the list is the first in the string representation)
    return [expr {([llength $x]-1)*$::math::bigfloat::segmentSize+[string length\
                [trimleft [lindex $x end]]]}]
}


################################################################################
# input : nombre (BigInt number)
# return : true if the number equals 0
################################################################################
proc ::math::bigfloat::intIsZero {nombre {raw 0}} {
    if {[isFloat $nombre]} {
        error "BigInt expected, received BigFloat instead"
    }
    if {!$raw} {
        set nombre [lrange $nombre 1 end]
    }
    while {wide([lindex $nombre end])==0} {
        # when you are at the last chunk...
        if {[llength $nombre]==1} {
            # ...the case is trivial : the number is equal to zero
            # if the last chunk is equal to zero
            if {wide([lindex $nombre 0])==0} {
                return 1
            }
            return 0
        }
        # iterate through the number from the beginning to end,
        # while the current chunk is equal to zero
        set nombre [lrange $nombre 0 end-1]
    }
    # we have found a chunk not equal to zero
    return 0
}

################################################################################
# right shift rounded by a number of digits lower than the size of a chunk
################################################################################
proc ::math::bigfloat::_roundshift {a n {raw 0}} {
    # fix the signed BigInt problem
    if {!$raw} {
        return [concat [lindex $a 0] [_roundshift [lrange $a 1 end] $n 1]]
    }
    set a [Dec $a 1]
    set filter %0.${::math::bigfloat::segmentSize}ld
    # We use format because it allows us to fetch whithout any error
    set dernier [format $filter [lindex $a 0]]
    set roundingValue [string index $dernier end-[expr {$n-1}]]
    # perform the right shift
    set result [_rshift $a $n 1]
    # we are rounding the BigInt
    if {$roundingValue>=5} {
        return [_rawIntAdd $result 1]
    }
    return $result
}



################################################################################
# right shift with upper rounding
################################################################################
proc ::math::bigfloat::roundshift {int {n 1} {raw 0}} {
    # dummy case
    if {$n==0} {return $int}
    if {[isFloat $int]} {
        error "BigInt required, received BigFloat instead"
    }
    # get done with signed BigInts
    if {!$raw} {
        set signe [lindex $int 0]
        set entier [roundshift [lrange $int 1 end] $n 1]
        # here : important point
        # when a number is right-shifted it may transit from negative to positive
        # (as long as zero is considered as positive)
        if {[rawIsNull $entier]} {
            set signe 1
        }
        return [concat $signe $entier]
    }
    if {[llength $int]==0} {
        return {0}
    }
    set int [Dec $int 1]
    set sSize $::math::bigfloat::segmentSize
    set smallShift [expr {$n%$sSize}]
    set bigShift [expr {$n/$sSize}]
    # res is a BigInt equal to int right shifted
    # by (($bigShift+1)*$::math::bigfloat::segmentSize) digits
    set res [lrange $int $bigShift end]
    set filter %0.${sSize}ld
    # if only shift by exactly N segments
    if {$smallShift==0} {
        incr bigShift -1
        # test to round the number
        if {[string index [format $filter [lindex $int $bigShift]] 0]>=5} {
            return [_rawIntAdd $res 1]
        }
        # rounded to floor
        return [Dec $res 1]
    }
    # smallShift > 0 & < 8 (chunksize)
    return [_roundshift $res $smallShift 1]
}



################################################################################
# right shift not rounded with a shift lower than the segment's size
################################################################################
proc ::math::bigfloat::_rshift {a n {raw 0}} {
    # treat the signed BigInt case
    if {!$raw} {
        return [concat [lindex $a 0] [_rshift [lrange $a 1 end] $n 1]]
    }
    set a [Dec $a 1]
    set result {}
    # a bit of formatting : padding left with zeros
    set filter %0.${::math::bigfloat::segmentSize}ld
    # the last chunk of the integer
    set lastChunk [format $filter [lindex $a 0]]
    # we need at least two chunks
    if {[llength $a]<2} {
        lappend a 0
    }
    set a [lrange $a 1 end]
    # the last but one chunk
    set lastButOne [format $filter [lindex $a 0]]
    # mixed is the candidate for being [lappend]ed
    # in mixed we put a mixture of the 2 last chunks
    set mixed [string range $lastButOne end-[expr {$n-1}] end]
    append mixed [string range $lastChunk 0 end-$n]
    set mixed [trimleft $mixed]
    while {[llength $a]>2} {
        lappend result $mixed
        # now we shift one chunk right and work on the 2 last chunks again
        # ... and again
        set lastChunk $lastButOne
        set a [lrange $a 1 end]
        set lastButOne [format $filter [lindex $a 0]]
        set mixed [string range $lastButOne end-[expr {$n-1}] end]
        append mixed [string range $lastChunk 0 end-$n]
        set mixed [trimleft $mixed]
    }
    # we desperatly need two chunks
    if {[llength $a]<2} {
        lappend a 0
    }
    lappend result $mixed
    set lastChunk $lastButOne
    set lastButOne [format $filter [lindex $a 1]]
    set mixed [string range $lastButOne end-[expr {$n-1}] end]
    append mixed [string range $lastChunk 0 end-$n]
    set mixed [trimleft $mixed]
    lappend result $mixed
    # there is only one chunk remaining and we shift it !
    lappend result [string range $lastButOne 0 end-$n]
    # result may contain heading zeros so let us normalize
    return [Dec $result 1]
}



################################################################################
# right shift without rounding
################################################################################
proc ::math::bigfloat::rshift {int {n 1} {raw 0}} {
    # dummy case
    if {$n==0} {return $int}
    if {[isFloat $int]} {
        error "BigInt required, received BigFloat instead"
    }
    if {!$raw} {
        set sign [lindex $int 0]
        set result [rshift [lrange $int 1 end] $n 1]
        # the result of the shifting of a negative integer
        # may be 0, which is positive
        if {[rawIsNull $result]} {
            set sign 1
        }
        # end of signed BigInt case
        return [concat $sign $result]
    }
    if {[llength $int]==0} {
        return {0}
    }
    set int [Dec $int 1]
    set l [length $int 1]
    if {$l<=$n} {
        return {0}
    }
    set sSize $::math::bigfloat::segmentSize
    set smallShift [expr {$n%$sSize}]
    set bigShift [expr {$n/$sSize}]
    # first shift by a multiple of segment's size (with lrange)
    set res [lrange $int $bigShift end]
    # if only shift by exactly "bigShift" segments
    if {$smallShift==0} {
        return $res
    }
    # then shift by a number of digits lower than segment's size
    set res [_rshift $res $smallShift 1]
    return $res
}


################################################################################
# compares two integers as with [string compare]
################################################################################
proc ::math::bigfloat::intCompare {a b {raw 0}} {
    set a [Dec $a $raw]
    set b [Dec $b $raw]
    if {$raw} {
        # A and B are raw : insert the sign to make signed BigInts of them
        set a [concat 1 $a]
        set b [concat 1 $b]
    }
    set signA [lindex $a 0]
    set signB [lindex $b 0]
    set la [llength $a]
    set lb [llength $b]
    if {$signA*$signB<0} {
        if {$signA>0} {
            # A > B
            return 1
        } else  {
            # A <B
            return -1
        }
    }
    # A and B have the same sign
    # when the chunk number are not equals, the case is clear :
    # if abs(a) < abs(b) : return -sign(a)
    if {$la<$lb} {return [expr {-1*$signA}]}
    # if abs(a) > abs(b) : return sign(a)
    if {$la>$lb} {return $signA}
    # now both numbers have the same chunk number
    # iterate from the heading to the ending of both BigInts
    # (the chunks are stored internally in reverse order, remember)
    set endA [lindex $a end]
    set endB [lindex $b end]
    incr la -1
    while {wide($endA)==wide($endB)} {
        incr la -1
        if {$la==0} {
            return 0
        }
        set endA [lindex $a $la]
        set endB [lindex $b $la]
    }
    # the same thing occurs with endA and endB
    # as above with the chunk numbers
    return [expr {(wide($endA)>wide($endB)?1:-1)*$signA}]
}



################################################################################
# adds n zeros at the end of $a
# left shift
################################################################################
proc ::math::bigfloat::lshift {a {n 1} {raw 0}} {
    if {[isFloat $a]} {
        error "BigInt required, received BigFloat instead"
    }
    # treat the case of signed BigInts
    if {!$raw} {
        return [concat [lindex $a 0] [lshift [lrange $a 1 end] $n 1]]
    }
    set sSize $::math::bigfloat::segmentSize
    set smallShift [expr {$n%$sSize}]
    set bigShift [expr {$n/$sSize}]
    # first : shift by a number lower than segment's size
    set a [_lshift $a $smallShift 1]
    # then : create a list of "0" chunks (to shift by a multiple of the segment's size)
    set l {}
    for {set i 0} {$i<$bigShift} {incr i} {
        lappend l 0
    }
    # and concatenate the chunks at the end of the BigInt (at the beginning of the list)
    set result [concat $l $a]
    if {[llength $result]==0} {
        return {0}
    }
    # normalize the result
    return [Dec $result 1]
}

################################################################################
# left shift by a number lower than chunk size
################################################################################
proc ::math::bigfloat::_lshift {a n {raw 0}} {
    # treat the signed BigInts
    if {!$raw} {
        return [concat [lindex $a 0] [_lshift [lrange $a 1 end] $n 1]]
    }
    # now a is an unsigned BigInt (a raw BigInt)
    set l [llength $a]
    if {$l==0} {
        return {0}
    }
    set m [expr {$n-1}]
    set result {}
    set carry [string repeat 0 $n]
    # filter for formatting with heading zeros
    set filter "%0.${::math::bigfloat::segmentSize}ld"
    # iterate from the end to the beginning of the number
    # (from the beginning to the end of the internal list representation)
    for {set i 0} {$i<$l} {incr i} {
        # format first to have a chunk at its full length
        set segment [format $filter [lindex $a $i]]
        # report the carry
        append segment $carry
        # set the new carry and append the current integer to the result
        set carry [string range $segment 0 $m]
        lappend result [trimleft [string range $segment $n end]]
    }
    return [concat $result $carry]
}

################################################################################
# trimleft for an integer stored as a string
# this proc is here mostly for the fact that integers beginning with a zero
# are treated as octal numbers, and produce an error
# when follows a decimal numbers
################################################################################
proc ::math::bigfloat::trimleft {number} {
    set number [string trimleft $number 0]
    # to keep the result as a valid number
    if {[string equal $number ""]} {
        return 0
    }
    return $number
}
################################################################################
# utility proc : get the sign of a signed BigInt
################################################################################
proc ::math::bigfloat::intSign {x} {
    if {[llength $x]>1} {
        return [lindex $x 0]
    }
    return [expr {($x<0)?-1:1}]
}




################################################################################
# conventions : raw means unsigned BigInt (without the leading sign)
# int mean BigInt, unsigned or signed
# underscore (_) procedures treat the second arguments as simple strings
# (so they are not longer than a segment (<= 8 digits)
# signe is the French name for sign
# entier is the French name for integer
################################################################################


################################################################################
# addition when b is small (not longer than a segment)
# this is done in a O(n) time, where n is the length of the greatest integer
################################################################################
proc ::math::bigfloat::_intAdd {a b} {
    set signeA [lindex $a 0]
    set a [lrange $a 1 end]
    if {[llength $a]==1} {
        # a and b are one segment's length
        set res [expr {$signeA*$a+$b}]
        set sign [expr {$res<0?-1:1}]
        set res [expr {abs($res)}]
        return [Dec [list $sign [expr {$res%$::math::bigfloat::segmentZero}] [expr {$res/$::math::bigfloat::segmentZero}]]]
    }
    set signeB [intSign $b]
    if {$signeA*$signeB<0} {
        # both signs are different : then call _rawIntSub
        if {$signeA<0} {
            return [concat $signeA [_rawIntSub $a $b]]
        } else  {
            return [concat 1 [_rawIntSub $a [expr {-$b}]]]
        }
    }
    # a and b do have the same sign, then call add
    # and with arguments : absolute values of both numbers
    set b [expr {[intSign $b]*$b}]
    return [concat $signeA [_rawIntAdd $a $b]]
}

################################################################################
# addition on unsigned ints when b is small (smaller than a half-wide)
################################################################################
proc ::math::bigfloat::_rawIntAdd {a b} {
    set segment $::math::bigfloat::segmentZero
    set a [Dec $a 1]
    # trimleft all zeros
    set b [trimleft $b]
    set result {}
    set segSize $::math::bigfloat::segmentSize
    while {[llength $a]>1} {
        # end of the number = first index
        # beginning of the number = last index
        set aEnd [lindex $a 0]
        # expr sum
        set sum [expr {$aEnd+$b}]
        set a [lrange $a 1 end]
        if {[string length $sum]<=$segSize} {
            # no carry : returns immediatly
            lappend result $sum
            set result [concat $result $a]
            return $result
        }
        set rest [expr {$sum%$segment}]
        set b [expr {$sum/$segment}]
        # b is the carry
        lappend result $rest
    }
    # now llength $a=1 so we can cast to integer
    set sum [expr {$a+$b}]
    # splitting the sum into two segments
    set remainder [expr {$sum%$segment}]
    set quotient [expr {$sum/$segment}]
    return [Dec [concat $result $remainder $quotient] 1]
}

################################################################################
# addition with unsigned integers
# the policy is to use the underscore adding proc the most
# when adding a BigInt by a simple integer the time is in O(n) and o(1)
################################################################################
proc ::math::bigfloat::rawIntAdd {a b} {
    set aEnd [lindex $a 0]
    set bEnd [lindex $b 0]
    # little is the sum of both last chunks
    set little [expr {$aEnd+$bEnd}]
    set result {}
    set segment $::math::bigfloat::segmentZero
    while {1} {
        # compute the carry ...
        set carry [expr {$little/$segment}]
        # and append the partial sum to the result
        lappend result [expr {$little%$segment}]
        if {[llength $a]>1} {
            set a [lrange $a 1 end]
        } else {
            # llength $a=1, so there is no more a
            # (we have already integrate the last chunk in the result)
            # llength $b unknown
            # call _rawIntAdd where the carry is the small integer
            if {[llength $b]==1} {
                set sum $carry
            } else  {
                set sum [_rawIntAdd [lrange $b 1 end] $carry]
            }
            return [Dec [concat $result $sum] 1]
        }
        if {[llength $b]>1} {
            set b [lrange $b 1 end]
        } else {
            # there is no more B, and A has still digits
            return [Dec [concat $result [_rawIntAdd $a $carry]] 1]
        }
        if {$carry != 0} {
            # we report the carry on A
            set a [_rawIntAdd $a $carry]
        }
        # end while loop by feeding addition
        set aEnd [lindex $a 0]
        set bEnd [lindex $b 0]
        set little [expr {$aEnd+$bEnd}]
    }
}

################################################################################
# addition for signed integers
################################################################################
proc ::math::bigfloat::intAdd {a b} {
    set a [Dec $a]
    set b [Dec $b]
    set signA [lindex $a 0]
    set signB [lindex $b 0]
    set a [lrange $a 1 end]
    set b [lrange $b 1 end]
    if {$signA*$signB<0} {
        # compare absolute values of $a and $b
        # 1 is here for raw version of intCompare
        if {[intCompare $a $b 1]>0} {
            # a BigInt is a list with a leading sign
            # and then integers in reverse form,
            # as the opposite of the string representation is.
            return [concat $signA [rawIntSub $a $b]]
            
        } else  {
            return [concat $signB [rawIntSub $b $a]]
        }
    }
    return [Dec [concat $signA [rawIntAdd $a $b]]]
}


################################################################################
# difference : returns a-b
################################################################################
proc ::math::bigfloat::intSub {a b} {
    # inverse $b, then add
    lset b 0 [expr {-[lindex $b 0]}]
    return [intAdd $a $b]
}

################################################################################
# tell if a raw BigInt (without sign) is null
################################################################################
proc ::math::bigfloat::rawIsNull {a} {
    foreach i $a {
        if {$i!=0} {
            return 0
        }
    }
    return 1
}

################################################################################
# difference between unsigned integers
################################################################################
proc ::math::bigfloat::rawIntSub {a b} {
    # if b>=a returns 0 (remember these are unsigned integers)
    if {[intCompare $a $b 1]<=0} {return 0}
    # dummy case
    if {[rawIsNull $b]} {return $a}
    # avoid artefacts due to nulls at the end of the list
    # (at the end of the list means at the beginning of the number)
    set a [Dec $a 1]
    set b [Dec $b 1]
    set result {}
    set carry 0
    set segmentMul $::math::bigfloat::segmentZero
    while {1} {
        # get the ends of the operands
        set aEnd [lindex $a 0]
        set bEnd [lindex $b 0]
        set diff [expr {$aEnd-$bEnd-$carry}]
        set carry 0
        if {wide($diff) < 0} {
            # set carry : 0-9=-9 carry 10-9=1
            set diff [expr {$segmentMul+$diff}]
            set carry 1
        }
        lappend result $diff
        if {[llength $a]>1} {
            set a [lrange $a 1 end]
        } else {
            # no more a, and a>b so we have not got any b more.
            return [Dec $result 1]
        }
        if {[llength $b]>1} {
            set b [lrange $b 1 end]
        } else {
            # no more b
            return [Dec [concat $result [_rawIntSub $a $carry]] 1]
        }
    }
}



################################################################################
# difference between a and b, when b is small
################################################################################
proc ::math::bigfloat::_rawIntSub {a b} {
    if {[llength $a]==1} {
        # we just have to add two normal integers
        set result [expr {$a-$b}]
        if {$result<0} {
            # beware : when substracting unsigned integers, we may round to zero
            # any negative number
            return {0}
        }
        return [Dec $result 1]
    }
    set segmentMul $::math::bigfloat::segmentZero
    # end of the number a (like when sub..ing with a pencil and some paper)
    set aEnd [lindex $a 0]
    set a [lrange $a 1 end]
    set diff [expr {$aEnd-$b}]
    set result {}
    if {$diff<0} {
        set diff [expr {$segmentMul+$diff}]
        # b was the second operand, so now it represents the carry
        # carry 1
        set b 1
    } else  {
        # diff>0 : carry null
        return [Dec [concat $result $diff $a] 1]
    }
    lappend result $diff
    # while there is a carry
    while {$b!=0} {
        if {[llength $a]==1} {
            # no more a, put the carry
            lappend result [expr {$a-$b}]
            # when b(carry)=0 trim this zero
            return [Dec $result 1]
        }
        set aEnd [lindex $a 0]
        set diff [expr {$aEnd-1}]
        if {$diff<0} {
            # carry 1
            # and add 10^9
            set diff [expr {$segmentMul+$diff}]
            set b 1
        } else  {
            # no carry
            set b 0
        }
        lappend result $diff
        set a [lrange $a 1 end]
        # no carry, so we can return the sum
        if {$b==0} {
            return [Dec [concat $result $a] 1]
        }
    }
}

################################################################################
# multiplies $a by $b
# a and b are unsigned BigInts
################################################################################
proc ::math::bigfloat::rawIntMul {a b} {
    if {[rawIsNull $a] || [rawIsNull $b]} {return 0}
    set result {0}
    set i 0
    set segment $::math::bigfloat::segmentZero
    set segSize $::math::bigfloat::segmentSize
    # should we use Karatsuba multiplication algorithm ?
    # are both numbers big enough?
    set threshold $::math::bigfloat::karatsubaLimit
    if {[llength $a]<$threshold && [llength $b]<$threshold} {
        # normal multiplication
        while {1} {
            set aEnd [lindex $a 0]
            set bEnd [lindex $b 0]
            # the product of two 9-digit integers may need to be considered
            # as a wide, for it to not throw an arithmetic overflow
            set last [expr {wide($aEnd)*wide($bEnd)}]
            # transforming $last into a list (raw BigInt)
            set last [list [expr {wide($last)%$segment}] [expr {wide($last)/$segment}]]
            if {[llength $a]>1 && [llength $b]>1} {
                set a [lrange $a 1 end]
                set b [lrange $b 1 end]
                # (A10^9+AEND) * (B10^9+BEND) = A*B*10^18 + 10^9*(A*BEND+B*AEND) + AEND*BEND
                # and then apply the same decomposition on A*B*10^18
                set comp [rawIntAdd [_rawIntMul $a $bEnd] [_rawIntMul $b $aEnd]]
                # the '1' argument of procedure lshift let us use raw BigInt
                set result [rawIntAdd $result [lshift [rawIntAdd $last [linsert $comp 0 0]] $i 1]]
                # $i is the parameter 10^18 in the formula 10^18*A*B
                # next time we will compute A*B and we have to keep trace of this parameter
                # (it needs a left shift of 18 digits
                incr i [expr {2*$segSize}]
            } else {
                if {[llength $a]>1} {
                    # no more b
                    set a [lrange $a 1 end]
                    return [rawIntAdd $result [lshift [rawIntAdd $last [linsert [_rawIntMul $a $bEnd]\
                            0 0]] $i 1]]
                }
                if {[llength $b]>1} {
                    # no more a
                    set b [lrange $b 1 end]
                    return [rawIntAdd $result [lshift [rawIntAdd $last [linsert [_rawIntMul $b $aEnd]\
                            0 0]] $i 1]]
                }
                return [rawIntAdd $result [lshift $last $i 1]]
            }
        }
    } else  {
        # begin Karatsuba algorithm
        # halfLen is the splitting point
        if {[intCompare $a $b 1]>0} {
            set halfLen [llength $a]
        } else  {
            set halfLen [llength $b]
        }
        # this line is taken from the bignum package of tcllib
        set halfLen [expr {($halfLen+($halfLen & 1))/2}]
        # splitting each operand in two equal-length numbers
        # A= A1*10^N+A2
        # B= B1*10^N+B2
        set A1 [lrange $a 0 end-$halfLen]
        set A2 [lrange $a end-[expr {$halfLen-1}] end]
        set B1 [lrange $b 0 end-$halfLen]
        set B2 [lrange $b end-[expr {$halfLen-1}] end]
        set A1 [Dec $A1 1]
        set B1 [Dec $B1 1]
        # algo. : A*B=A1*B1*10^(2N) + (A1*B2+A2*B1)*10^N + A2*B2
        set product1 [rawIntMul $A1 $B1]
        set product2 [rawIntMul $A2 $B2]
        # compute (A2+A1)*(B2+B1)
        # the middle term becomes product3-(product1+product2)
        set product3 [rawIntMul [rawIntAdd $A1 $A2] \
                [rawIntAdd $B1 $B2]]
        set product3 [rawIntSub $product3 [rawIntAdd $product1 $product2]]
        set product1 [lshift $product1 [expr {$halfLen*2*$::math::bigfloat::segmentSize}] 1]
        set product3 [lshift $product3 [expr {$halfLen*$::math::bigfloat::segmentSize}] 1]
        return [rawIntAdd $product1 [rawIntAdd $product3 $product2]]
        # END of Karatsuba algorithm
    }
}

################################################################################
# _rawIntMul : multiplication for an unsigned BigNum and an unsigned integer
# time is in O(n)
################################################################################
proc ::math::bigfloat::_rawIntMul {a b} {
    if {$b==0||[rawIsNull $a]} {return 0}
    set result {}
    set carry {0}
    set segment $::math::bigfloat::segmentZero
    while {1} {
        if {[llength $a]>1} {
            set aEnd [lindex $a 0]
            set a [lrange $a 1 end]
        } else {
            # 'a' is a list of one unsigned integer,
            # so list->integer cast is possible
            set little [expr {wide($a)*wide($b)}]
            set little [list [expr {wide($little)%$segment}] [expr {wide($little)/$segment}]]
            return [Dec [concat $result [rawIntAdd $carry $little]] 1]
        }
        set little [expr {wide($aEnd)*wide($b)}]
        # accumulate the carry
        # first split the current product into two segments
        set carryNext [expr {wide($little)/$segment}]
        set little [expr {wide($little)%$segment}]
        # distribute the accumulated carry (which is more an accumulator than a carry)
        # and the last product
        set carry [_rawIntAdd $carry $little]
        if {[llength $carry]>1} {
            # the carry might be longer than one segment, so we append the last segment
            lappend result [lindex $carry 0]
            # and we accumulate
            set carry [_rawIntAdd [lrange $carry 1 end] $carryNext] 
        } else {
            # do the same : prepend the carry at the beginning of the result...
            set result [concat $result $carry]
            # ... and feed the carry for the next loop
            set carry [list $carryNext]
        }
        # when there is no more A to feed the operation, end it.
        if {[rawIsNull $a]} {return [Dec [concat $result $carry] 1]}
    }
}

################################################################################
# _intMul : A times B where :
# A is a BigInt
# B is an integer
################################################################################
proc ::math::bigfloat::_intMul {x y} {
    # not necessary to normalize
    #set x [Dec $x]
    # store the sign
    set sign [expr {[intSign $x]*[intSign $y]}]
    set x [lrange $x 1 end]
    # call the raw equivalent procedure
    set x [_rawIntMul $x [expr {abs($y)}]]
    # retrieve the sign
    return [concat $sign $x]
}

################################################################################
# intMul : BigInt multiplication
################################################################################
proc ::math::bigfloat::intMul {a b} {
    #set a [Dec $a]
    #set b [Dec $b]
    set signA [lindex $a 0]
    set signB [lindex $b 0]
    return [concat [expr {$signA*$signB}] [rawIntMul [lrange $a 1 end]\
            [lrange $b 1 end]]]
}



################################################################################
# returns 2 at the exp'th power (exp is a simple integer)
################################################################################
proc ::math::bigfloat::powOfTwo {exp} {
    set res {1}
    set a {2}
    while {1} {
        # decomposition in binary base of the exponent :
        # exponent=(Product(i) e_i*2^i)
        set reste [expr {$exp%2}]
        set exp [expr {$exp/2}]
        if {$reste!=0} {
            # mul res by 2^i
            set res [rawIntMul $res $a]
        }
        if {$exp==0} {
            # no more exponent, return the power of two by the exponent
            return $res
        }
        # compute 2^i
        set a [rawIntMul $a $a]
    }
}

################################################################################
# a new division algorithm : divide digit by digit as by hand
################################################################################
proc ::math::bigfloat::newRawIntDivqr {num denom} {
    set num [Dec $num 1]
    set denom [Dec $denom 1]
    set l [length $num 1]
    incr l -[length $denom 1]
    incr l
    set result {0}
    while {$l>=0} {
        # take a middle value and a step for having an (1,9) range
        set test 5
        set step 4
        set cmp [intCompare [lshift [_rawIntMul $denom $test] $l 1] $num 1]
        while {$cmp!=0 && $step!=0} {
            # perform dichotomy : add or sub a step to the test value
            if {$cmp>0} {
                incr test -$step
            } else {
                incr test $step
            }
            # when 1*denom>num, next value of test is : -1 so it should be a zero
            if {$test<0} {
                set test 0
                break
            }
            # when 9*denom<num, next value of test is : 11 so it should be nine
            if {$test>9} {
                set test 9
                break
            }
            # set the next step for dichotomy
            set step [expr {$step/2}]
            # compute test*denom compared to num
            set cmp [intCompare [lshift [_rawIntMul $denom $test] $l 1] $num 1]
        }
        # the digit is '0' so nothing to do : continue to loop
        if {$test==0} {
            incr l -1
            continue
        }
        # if test*denom>num : decrement test
        if {$cmp>0} {
            incr test -1
        }
        # if test*denom<num && interval is 0
        # example : 5*denom<0 -> 9*denom>0 -> 7*denom>0 (step=1)
        # -> 6*denom<0 (step=0)
        # test is the true digit
        set result [rawIntAdd $result [lshift [list $test] $l 1]]
        set num [rawIntSub $num [lshift [_rawIntMul $denom $test] $l 1]]
        incr l -1
    }
    # result contains the quotient, and num the remainder
    return [list $result $num]
}

################################################################################
# the new division algorithm : slower than the old one on small numbers,
# but asymptotically faster (it beats the old one on big numbers)
# Algorithm taken from math::bignum tcllib's package. (works like
# a binary one)
# not used anymore : replaced by newRawIntDivqr
################################################################################
proc ::math::bigfloat::newBinIntDivqr {n d} {
    set n [Dec $n 1]
    set d [Dec $d 1]
    # we have N digits (in decimal) -> so we have log(10)/log(2)*(N+1) binary digits
    set bit [expr {int(ceil(log(10)*([length $n 1]+1)/log(2)))+1}]
    set r {0}
    set q {0}
    set refPower [powOfTwo $bit]
    # for i=bit downto 0 (refPower=2^i)
    while {$bit >= 0} {
        set r [_rawIntMul $r 2]
        # let n_i be the binary digit of n at index i
        # if n_i=1, add 1 to remainder and set n_i to 0
        if {[intCompare $n $refPower 1]>=0} {
            set r [_rawIntAdd $r 1]
            set n [rawIntSub $n $refPower]
        }
        # when the remainder is greater than the divider, increment the quotient
        # and substract the divider to the remainder
        if {[intCompare $r $d 1] >= 0} {
            set r [rawIntSub $r $d]
            set q [rawIntAdd $q $refPower]
        }
        # the following line faster than : decrement bit, then compute 2^bit
        set refPower [_rawIntDiv $refPower 2]
        incr bit -1
    }
    return [list $q $r]
}


################################################################################
# division for unsigned BitInt
# returns a list containing quotient and rest
################################################################################
proc ::math::bigfloat::rawIntDivqr {a b} {
    set a [Dec $a 1]
    set b [Dec $b 1]
    set lb [llength $b]
    if {[llength $a]>4 && $lb>[llength $a]/3} {
        # for big integers, that version is faster
        return [newRawIntDivqr $a $b]
    }
    if {$lb==1} {
        set b [lindex $b 0]
        # division when $b is small
        return [_rawIntDivqr $a $b]
    }
    set cmp [intCompare $a $b 1]
    # some trivial cases treated to avoid painful computations
    if {$cmp<0} {return [list {0} $a]}
    if {$cmp==0} {return [list {1} {0}]}
    # a/(b+c.10^8)=(a/b)/(1+c.10^8/b)
    # a_0=a/b
    # b_0=1+c.10^8/b
    # a_n+1 = a_n/(b_n % 10^8)
    # b_n+1 = 1+(b_n / 10^8).10^8/(b_n % 10^8)
    # we will compute this way an interval in which the quotient is surely
    # ATTENTION ! Take care when b_n % 10^8 equals 1 !!!
    set amin $a
    set amax $a
    set bmin $b
    set bmax $b
    #puts "first fit"
    # upper limit : amax/cmin
    while {[llength $bmin]>1} {
        set mod [lindex $bmin 0]
        if {$mod>1} {
            # b_n / 10^8
            lset bmin 0 0
            set bmin [_rawIntAdd [_rawIntDiv $bmin $mod] 1]
            set amax [_rawIntDiv $amax $mod]
        } else {
            # include the two cases : mod=1 & mod=0
            set bmin [lrange $bmin 1 end]
            set amax [_rawIntAdd [lrange $amax 1 end] 1]
        }
    }
    set qmax [_rawIntDiv $amax [lindex $bmin 0]]
    #puts "second fit"
    # the next while does the same thing but with a floor value
    while {[llength $bmax]>1} {
        set mod [lindex $bmax 0]
        if {$mod==1} {
            # if we let $mod as 1, it would loop endlessly
            lset bmax 0 0
            set bmax [_rawIntAdd $bmax 2]
            set amin [_rawIntDiv $amin 2]
        } elseif {$mod!=0} {
            # b_n / 10^8
            lset bmax 0 0
            set bmax [_rawIntAdd [_rawIntDiv $bmax $mod] 2]
            set amin [_rawIntDiv $amin [expr {$mod+1}]]
        } else {
            set bmax [lrange $bmax 1 end]
            set amin [lrange $amin 1 end]
        }
    }
    set qmin [_rawIntDiv $amin [expr {[lindex $bmax 0]+1}]]
    # qmin and qmax are forming a quotient knowledge interval
    # pmin and pmax are products of the quotient by the divider
    set pmin [rawIntMul $qmin $b]
    set pmax [rawIntMul $qmax $b]
    # rmin and rmax are forming a remainder knowledge interval
    set rmin [rawIntSub $a $pmin]
    set rmax [rawIntSub $a $pmax]
    if {[intCompare $pmin $a 1]<=0 && [intCompare $rmin $b 1]<0} {
        return [list $qmin $rmin]
    }
    if {[intCompare $pmax $a 1]<=0 && [intCompare $rmax $b 1]<0} {
        return [list $qmax $rmax]
    }
    #puts "before dichotomy"
    # we use dichotomy to find the quotient (and at the same time, the remainder)
    while {1} {
        # dichotomy to be redone
        # q = (qmin+qmax)/2
        set q [_rawIntDiv [rawIntAdd $qmin $qmax] 2]
        if {[intCompare $q $qmin 1]==0} {
            if {[intCompare [_rawIntAdd $q 1] $qmax 1]>=0} {
                return [list $q [rawIntSub $a [rawIntMul $b $q]]]
            }
        }
        # if q*b>a (q>a/b)
        set p [rawIntMul $q $b]
        if {[intCompare $p $a 1]>0} {
            # q is the upper bound
            set qmax $q
            continue
        }
        # q*b<a so rest=a-b*q and rest>0
        set r [rawIntSub $a $p]
        if {[intCompare $r $b 1]>=0} {
            # rest>b, then qmin=q
            set qmin $q
            continue
        }
        # q*b<a and rest<b : we can return
        return [list $q $r]
    }
}


################################################################################
# division with :
# A : an unsigned BigInt
# B : an unsigned integer no longer than one segment
# returns a list : 1. the quotient 2. the remainder
################################################################################
proc ::math::bigfloat::_rawIntDivqr {a b} {
    set result [list 0]
    set segMul $::math::bigfloat::segmentZero
    set a [Dec $a 1]
    set b [trimleft $b]
    if {wide($b)==0} {
        error "divide by zero"
    }
    # used to pad number segments with zeros
    set filter "%0.${::math::bigfloat::segmentSize}ld"
    while {[llength $a]>2} {
        # beginning of A : end of the list
        # take the two beginning segments and convert them to a wide integer 
        set debuta [trimleft [lindex $a end][format $filter [lindex $a end-1]]]
        set a [lrange $a 0 end-2]
        # compute the quotient and the rest of the partial division
        set quotient [expr {wide($debuta)/$b}]
        set reste [expr {wide($debuta)%$b}]
        # convert quotient back to a raw BigInt
        set quotient [list [expr {wide($quotient)%$segMul}] [expr {wide($quotient)/$segMul}]]
        set quotient [lshift $quotient [expr {$::math::bigfloat::segmentSize*[llength $a]}] 1]
        # take account of the quotient and the rest
        set result [rawIntAdd $result $quotient]
        set a [concat $a $reste]
    }
    # do the same thing with [llength $a]=1 or 2
    if {[llength $a]==2} {
        set a [trimleft [lindex $a end][format $filter [lindex $a end-1]]]
    } else  {
        set a [lindex $a 0]
    }
    # perform the division (the rest is the remainder of the entire operation
    set quotient [expr {wide($a)/$b}]
    return [list [rawIntAdd $result [list [expr {wide($quotient)%$segMul}] [expr {wide($quotient)/$segMul}]]] [expr {wide($a)%$b}]]
}

################################################################################
# return the quotient of the division of A by B
# A : a raw BigInt
# B : a normal integer with less than 9 digits
################################################################################
proc ::math::bigfloat::_rawIntDiv {a b} {
    return [lindex [_rawIntDivqr $a $b] 0]
}



################################################################################
# intDivqr : returns quotient and rest of a BigInt division
################################################################################
proc ::math::bigfloat::intDivqr {a b} {
    set signeA [lindex $a 0]
    set signeB [lindex $b 0]
    set a [lrange $a 1 end]
    set b [lrange $b 1 end]
    foreach {quotient rest} [rawIntDivqr $a $b] {break}
    # the rest has to be of the same sign as the divider
    # if both signs are different, one should apply a patch
    set quotient [Dec $quotient 1]
    set rest [Dec $rest 1]
    if {$signeB<0} {
        if {$signeA<0} {
            return [list [concat 1 $quotient] [concat -1 $rest]]
        } else {
            # patch to make the rest be negative
            set quotient [_rawIntAdd $quotient 1]
            set rest [rawIntSub $b $rest]
            return [list [concat -1 $quotient] [concat -1 $rest]]
        }
    }
    if {$signeA<0} {
        # patch to make the rest be positive
        set quotient [_rawIntAdd $quotient 1]
        set rest [rawIntSub $b $rest]
        return [list [concat -1 $quotient] [concat 1 $rest]]
    }
    # nothing more to do, because both signs are positive
    return [list [concat 1 $quotient] [concat 1 $rest]]
}

################################################################################
# _intDiv : divide a BigInt A by an integer B
################################################################################
proc ::math::bigfloat::_intDiv {a b} {
    set a [Dec $a]
    set signe [lindex $a 0]
    return [concat $signe [_rawIntDiv [lrange $a 1 end] $b]]
}

################################################################################
# returns round(A/B)
# input : a & b (BigInts)
################################################################################
proc ::math::bigfloat::rnddiv {a b} {
    foreach {quotient rest} [intDivqr $a $b] {break}
    # when the rest >= 2b, the quotient is xxx.5xxxx -> so we have to increment it
    if {[intCompare $rest [_intDiv $b 2]]>=0} {
        return [_intAdd $quotient 1]
    }
    return $quotient
}

################################################################################
# returns the quotient incremented
################################################################################
proc ::math::bigfloat::incrdiv {a b} {
    set q [intDiv $a $b]
    return [_intAdd $q 1]
}

################################################################################
# returns the quotient
################################################################################
proc ::math::bigfloat::intDiv {a b} {
    return [lindex [intDivqr $a $b] 0]
}


################################################################################
# returns the rest of the division
################################################################################
proc ::math::bigfloat::intMod {a b} {
    return [lindex [intDivqr $a $b] 1]
}


################################################################################
# checks that n is a BigInt ; returns 0 when it is a BigFloat, 1 elswhere
# a BigInt is supposed to be a list of :
# index 0 : the sign (1 or -1)
# index 1..n : the number in reverse order, split into segments of 9 digits
# a BigInt is assumed to [llength] at least 2
################################################################################
proc ::math::bigfloat::isInt {n} {
    if {[llength $n]<2} {
        return 0
    }
    set sign [lindex $n 0]
    if {[string equal $sign 1]||[string equal $sign -1]} {
        return 1
    }
    return 0
}


################################################################################
# PART II
# procedures that handle floating-point numbers
# they are sorted by name (after eventually removing the underscores)
# BigFloats are internally represented as a list :
# {"F" Mantissa Exponent Delta} where "F" is a character which determins
# the datatype, Mantissa and Delta are two BigInt and Exponent a simple integer.
#
# The BigFloat value equals to (Mantissa +/- Delta)*10^Exponent
# Example : [fromstr 1.00] -> {F {1 100} -2 {1 1}}
# [fromstr -4e-10 10] -> {F {-1 0 400} -10 {1 1}} (see the manual for explainations)
# When calling fromstr, the Delta parameter is always set to 1.
# When you perform some computations, the Delta parameter (uncertainty) may increase.
# It is kept at 2 digits at most with the normalize procedure.
# Retrieving the parameters of a BigFloat is often done with that command :
# foreach {dummy int exp delta} $bigfloat {break}
# (dummy is not used, it is just used to get the "F".
# The isInt and checkFloat procedures are using this character to check whether the
# number is a BigFloat or not. If not, the isInt procedure assumes that it is a BigInt
################################################################################


################################################################################
# returns the absolute value
################################################################################
proc ::math::bigfloat::abs {number} {
    checkNumber number
    if {[isInt $number]} {
        # set sign to 1 for a BigInt
        lset number 0 1
        return $number
    }
    # set sign to 1 for a BigFloat
    lset number 1 0 1
    return $number
}


################################################################################
# arccosinus of a BigFloat
################################################################################
proc ::math::bigfloat::acos {x} {
    # handy proc for checking datatype
    checkFloat x
    foreach {dummy entier exp delta} $x {break}
    set precision [expr {($exp<0)?(-$exp):1}]
    # acos(0.0)=Pi/2
    if {[isnull $x]} {
        return [div [pi $precision] {1 2}]
    }
    # acos(-x)= Pi/2 + asin(x)
    if {[intSign $entier]<0} {
        set pi_sur_deux [div [pi $precision] {1 2}]
        return [add $pi_sur_deux [asin [abs $x]]]
    }
    # we always use _asin to compute the result
    # but as it is a Taylor development, the value given to [_asin]
    # has to be a bit smaller than 1 ; by using that trick : acos(x)=asin(sqrt(1-x^2))
    # we can limit the entry of the Taylor development below 1/sqrt(2)
    if {[compare $x {F {1 7071} -4 {1 1}}]>0} {
        # x > sqrt(2)/2 : trying to make _asin converge quickly 
        # creating 0 and 1 with the same precision as the entry
        set zero [list F {1 0} -$precision {1 1}]
        set un [list F [lshift {1 1} $precision] -$precision {1 1}]
        # acos(1.0)=0.0
        if {[equal $un $x]} {
            return $zero
        }
        if {[compare $un $x]<0} {
            # the behavior assumed because acos(x) is not defined
            # when |x|>1
            error "acos on a number greater than 1"
        }
        # acos(x) = asin(sqrt(1 - x^2))
        # since 1 - cos(x)^2 = sin(x)^2
        # x> sqrt(2)/2 so x^2 > 1/2 so 1-x^2<1/2
        set x [sqrt [sub $un [mul $x $x]]]
        # the parameter named x is smaller than sqrt(2)/2
        return [_asin $x]
    }
    set pi_sur_deux [div [pi $precision] {1 2}]
    # acos(x) = Pi/2 - asin(x)
    # x<sqrt(2)/2 here too
    return [sub $pi_sur_deux [_asin $x]]
}


################################################################################
# returns A + B
################################################################################
proc ::math::bigfloat::add {a b} {
    checkNumber a b
    if {[isInt $a]} {
        if {[isInt $b]} {
            # intAdd adds two BigInts
            return [intAdd $a $b]
        }
        # adds the BigInt a to the BigFloat b
        return [addInt2Float $b $a]
    }
    if {[isInt $b]} {
        # ... and vice-versa
        return [addInt2Float $a $b]
    }
    # retrieving parameters from A and B
    # entier means integer in French
    foreach {dummy entierA expA deltaA} $a {break}
    foreach {dummy entierB expB deltaB} $b {break}
    # when a number has more precision than the other (for example,
    # when A=0.001 and B=2.01, we need to put the other at the decimal level
    # that is why we shift left the number which has the greater exponent
    if {$expA>$expB} {
        set diff [expr {$expA-$expB}]
        set entierA [lshift $entierA $diff]
        set deltaA [lshift $deltaA $diff]
        set entierA [intAdd $entierA $entierB]
        set deltaA [intAdd $deltaA $deltaB]
        return [normalize [list F $entierA [expr {$expA-$diff}] $deltaA]]
    } elseif {$expA==$expB} {
        # nothing to shift left
        return [normalize [list F [intAdd $entierA $entierB] \
                $expA [intAdd $deltaA $deltaB]]]
    } else {
        set diff [expr {$expB-$expA}]
        set entierB [lshift $entierB $diff]
        set deltaB [lshift $deltaB $diff]
        set entierB [intAdd $entierA $entierB]
        set deltaB [intAdd $deltaB $deltaA]
        return [normalize [list F $entierB [expr {$expB-$diff}] $deltaB]]
    }
}

################################################################################
# returns the sum A(BigFloat) + B(BigInt)
# the greatest advantage of this method is that the uncertainty
# of the result remains unchanged, in respect to the entry's uncertainty (deltaA)
################################################################################
proc ::math::bigfloat::addInt2Float {a b} {
    # type checking
    checkFloat a
    if {![isInt $b]} {
        error "'$b' is not a BigInt"
    }
    # retrieving data from $a
    foreach {dummy entierA expA deltaA} $a {break}
    # to add an int to a BigFloat,...
    if {$expA>0} {
        # we have to put the integer entierA
        # to the level of zero exponent : 1e8 --> 100000000e0
        set entierA [lshift $entierA $expA]
        set deltaA [lshift $deltaA $expA]
        set entierA [intAdd $entierA $b]
        # we have to normalize, because we have shifted the mantissa
        # and the uncertainty left
        return [normalize [list F $entierA 0 $deltaA]]
    } elseif {$expA==0} {
        # entierA is already at integer level : float=(entierA)e0
        return [normalize [list F [intAdd $entierA $b] \
                0 $deltaA]]
    } else {
        # here we have something like 234e-2 + 3
        # we have to shift the integer left by the exponent |$expA|
        set b [lshift $b [expr {-$expA}]]
        set entierA [intAdd $entierA $b]
        return [normalize [list F $entierA $expA $deltaA]]
    }
}


################################################################################
# arcsinus of a BigFloat
################################################################################
proc ::math::bigfloat::asin {x} {
    # type checking
    checkFloat x
    foreach {dummy entier exp delta} $x {break}
    set precision [expr {($exp<0)?(-$exp):1}]
    # when x=0, return 0 at the same precision as the input was
    if {[isnull $x]} {
        return [list F {1 0} -$precision {1 1}]
    }
    # asin(-x)=-asin(x)
    if {[intSign $entier]<0} {
        return [opp [asin [abs $x]]]
    }
    # now a little trick : asin(x)=Pi/2-asin(sqrt(1-x^2))
    # so we can limit the entry of the Taylor development
    # to 1/sqrt(2)~0.7071
    # the comparison is : if x>0.7071 then ...
    if {[compare $x {F {1 7071} -4 {1 1}}]>0} {
        set pi_sur_deux [div [pi $precision] {1 2}]
        set un [list F [lshift {1 1} $precision] -$precision {1 1}]
        # asin(1)=Pi/2 (with the same precision as the entry has)
        if {[equal $un $x]} {
            return $pi_sur_deux
        }
        if {[compare $x $un]>0} {
            error "asin on a number greater than 1"
        }
        set x [sqrt [sub $un [mul $x $x]]]
        return [sub $pi_sur_deux [_asin $x]]
    }
    return [normalize [_asin $x]]
}

################################################################################
# _asin : arcsinus of numbers between 0 and +1
################################################################################
proc ::math::bigfloat::_asin {x} {
    # Taylor development
    # asin(x)=x + 1/2 x^3/3 + 3/2.4 x^5/5 + 3.5/2.4.6 x^7/7 + ...
    foreach {dummy mantissa exp delta} $x {break}
    set precision [expr {-$exp}]
    if {$precision+1<[length $mantissa]} {
        error "sinus greater than 1"
    }
    # precision is the number of after-dot digits
    set result $mantissa
    set delta_final $delta
    # resultat is the final result, and delta_final
    # will contain the uncertainty of the result
    # square is the square of the mantissa
    # we have to shift right because :
    # if mantissa is 230 and precision = 3
    # x=0.230 and x^2=230*230*10^-6
    # the square of the mantissa would be 230*230=52900
    # but precision=3 so it would be : 52.900 instead of : 0.052
    # so we have to shift right by 3 digits
    # square=mantissa^2 >> 3 (decimal shift, equivalent to mul by 10^-3
    set square [rshift [intMul $mantissa $mantissa] $precision]
    # dt is the uncertainty
    set dt [rshift [_intMul [intMul $mantissa $delta] 2] $precision]
    # these three are required to compute the fractions implicated into
    # the development (of Taylor, see former)
    set num 1
    set i 3
    set denom 2
    set delta [intAdd [intMul $delta $square] [intMul $dt $mantissa]]
    set delta [rshift [_intDiv [_intMul $delta $num] $denom] $precision]
    set mantissa [rshift [intMul $mantissa $square] $precision]
    set mantissa [_intDiv $mantissa $denom]
    set mantissa_temp [_intDiv $mantissa $i]
    set delta_temp [_intDiv $delta $i]
    while {[intCompare $mantissa_temp $delta_temp]>0} {
        set result [intAdd $result $mantissa_temp]
        set delta_final [intAdd $delta_final $delta_temp]
        incr num 2
        incr i 2
        incr denom 2
        set delta [intAdd [intMul $delta $square] [intMul $dt $mantissa]]
        set delta [rshift [_intDiv [_intMul $delta $num] $denom] $precision]
        set mantissa [rshift [intMul $mantissa $square] $precision]
        set mantissa [_intDiv [_intMul $mantissa $num] $denom]
        set mantissa_temp [_intDiv $mantissa $i]
        set delta_temp [_intDiv $delta $i]
    }
    return [list F $result $exp $delta_final]
}

################################################################################
# arctangent : returns atan(x)
################################################################################
proc ::math::bigfloat::atan {x} {
    checkFloat x
    foreach {dummy mantissa exp delta} $x {break}
    set precision [expr {($exp<0)?-$exp+1:1}]
    # atan(0)=0
    if {[isnull $x]} {
        return [list F {1 0} -$precision {1 1}]
    }
    # atan(-x)=-atan(x)
    if {[intSign $mantissa]<0} {
        return [opp [atan [abs $x]]]
    }
    # now x is strictly positive
    # at this moment, we are trying to limit |x| to a fair acceptable number
    # to ensure that Taylor development will converge quickly
    set un [list F [lshift {1 1} $precision] -$precision {1 1}]
    if {[compare $un $x]<0} {
        # compare x to 2.4142
        if {[compare $x {F {1 24142} -4 {1 1}}]<0} {
            # atan(x)=Pi/4 + atan((x-1)/(x+1))
            # as 1<x<2.380 : (x-1)/(x+1)=1-2/(x+1) belongs to
            # the range :  ]0,1-2/3.414[
            # that equals  ]0,0.414[
            set pi_sur_quatre [div [pi $precision] {1 4}]
            return [add $pi_sur_quatre [atan \
                    [div [sub $x $un] [add $x $un]]]]
        }
        # atan(x)=Pi/2-atan(1/x)
        # 1/x < 1/2.414 so the argument is lower than 0.414
        set pi_sur_deux [div [pi $precision] {1 2}]
        return [sub $pi_sur_deux [atan [div $un $x]]]
    }
    if {[compare $x {F {1 4142} -4 {1 1}}]>0} {
        # atan(x)=Pi/4 + atan((x-1)/(x+1))
        # x>0.420 so (x-1)/(x+1)=1 - 2/(x+1) > 1-2/1.414
        #                                    > -0.414
        # x<1 so (x-1)/(x+1)<0
        set pi_sur_quatre [div [pi $precision] {1 4}]
        return [add $pi_sur_quatre [atan \
                [div [sub $x $un] [add $x $un]]]]
    }
    set precision [expr {-$exp}]
    # precision increment : to have less uncertainty
    # we add a little more precision so that the result would be more accurate
    set n [expr {int($precision*log(10)/log(42)+1)}]
    incr precision $n
    set mantissa [lshift $mantissa $n]
    set delta [lshift $delta $n]
    # end of adding precision increment
    # now computing Taylor development :
    # atan(x)=x - x^3/3 + x^5/5 - x^7/7 ... + (-1)^n*x^(2n+1)/(2n+1)
    set result $mantissa
    set delta_fin $delta
    # we store the square of the integer (mantissa)
    set delta_square [_intMul $delta 2]
    set square [rshift [intMul $mantissa $mantissa] $precision]
    # the (2n+1) divider
    set div 3
    # computing precisely the uncertainty
    set delta [_intAdd [rshift [intAdd [intMul $delta_square $mantissa] \
            [intMul $delta $square]] $precision] 1]
    # temp contains (-1)^n*x^(2n+1)
    set temp [opp [rshift [intMul $mantissa $square] $precision]]
    set t [_intDiv $temp $div]
    set dt [_intAdd [_intDiv $delta $div] 1]
    while {![intIsZero $t]} {
        set result [intAdd $result $t]
        set delta_fin [intAdd $delta_fin $dt]
        incr div 2
        set delta [_intAdd [rshift [intAdd [intMul $delta_square [abs $temp]] \
                [intMul $delta $square]] $precision] 1]
        set temp [opp [rshift [intMul $temp $square] $precision]]
        set t [_intDiv $temp $div]
        set dt [_intAdd [_intDiv $delta $div] 1]
    }
    # we have to normalize because the uncertainty might be greater than 99
    # moreover it is the most often case
    return [normalize [list F $result [expr {$exp-$n}] $delta_fin]]
}

################################################################################
# compute atan(1/integer) at a given precision
# this proc is only used to compute Pi
# it is using the same Taylor development as [atan]
################################################################################
proc ::math::bigfloat::_atanfract {integer precision} {
    set n [expr {int($precision*log(10)/log(25)+1)}]
    set integer [list 1 $integer]
    incr precision $n
    set un [lshift {1 1} $precision]
    # first term of the development : 1/integer
    set a [intDiv $un $integer]
    # 's' will contain the result
    set s $a
    # integer^2
    set square [intMul $integer $integer]
    set denom 3
    # $t is (-1)^n*x^(2n+1)
    set t [opp [intDiv $a $square]]
    set u [_intDiv $t $denom]
    # we break the loop when the current term of the development is null
    while {![intIsZero $u]} {
        set s [intAdd $s $u]
        # denominator= (2n+1)
        incr denom 2
        set t [opp [intDiv $t $square]]
        set u [_intDiv $t $denom]
    }
    # go back to the initial precision
    return [rshift $s $n]
}

################################################################################
# returns the integer part of a BigFloat, as a BigInt
# the result is the same one you would have
# if you had called [expr {ceil($x)}]
################################################################################
proc ::math::bigfloat::ceil {number} {
    checkFloat number
    set number [normalize $number]
    if {[isnull $number]} {
        return {1 0}
    }
    foreach {dummy mantissa exp delta} $number {break}
    if {$exp>0} {
        error "not enough precision to round (in ceil)"
    }
    set decimal [lshift {1 1} [expr {-$exp}]]
    foreach {mantissa last} [intDivqr $mantissa $decimal] {break}
    if {[intSign $mantissa]<0} {
        # mantissa's ceiling is mantissa+1
        return [_intAdd $mantissa 1]
    }
    # if sign<0 : nothing to do but to return integer part
    # if sign>0 :
    #                - if last=0 do nothing
    #                - if last<>0 increment the integer
    if {[intSign $mantissa]>0 && [intCompare [abs $last] $delta]>0} {
        set mantissa [_intAdd $mantissa 1]
    }
    return $mantissa
}


################################################################################
# checks each variable to be a BigFloat
# arguments : each argument is the name of a variable to be checked
################################################################################
proc ::math::bigfloat::checkFloat {args} {
    foreach x $args {
        upvar $x n
        if {![isFloat $n]} {
            error "BigFloat expected : received '$n'"
        }
    }
}

################################################################################
# checks if each number is either a BigFloat or a BigInt
# arguments : each argument is the name of a variable to be checked
################################################################################
proc ::math::bigfloat::checkNumber {args} {
    foreach i $args {
        upvar $i x
        if {![isInt $x] && ![isFloat $x]} {
            error "'$x' is not a number"
        }
    }
}


################################################################################
# returns 0 if A and B are equal, else returns 1 or -1
# accordingly to the sign of (A - B)
################################################################################
proc ::math::bigfloat::compare {a b} {
    if {[isInt $a] && [isInt $b]} {
        return [intCompare $a $b]
    }
    checkFloat a b
    if {[equal $a $b]} {return 0}
    return [lindex [sub $a $b] 1 0]
}




################################################################################
# gets cos(x)
# throws an error if there is not enough precision on the input
################################################################################
proc ::math::bigfloat::cos {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    if {$exp>-2} {
        error "not enough precision on floating-point number"
    }
    set precision [expr {-$exp}]
    # cos(2kPi+x)=cos(x)
    foreach {n integer} [divPiQuarter $integer $precision] {break}
    # now integer>=0 and <Pi/2
    set d [tostr [intMod $n {1 4}]]
    # add trigonometric circle turns number to delta
    set delta [intAdd [abs $n] $delta]
    set signe 1
    # cos(Pi-x)=-cos(x)
    # cos(-x)=cos(x)
    switch -- $d {
        1 {set signe -1;set l [_sin2 $integer $precision $delta]}
        2 {set signe -1;set l [_cos2 $integer $precision $delta]}
        0 {set l [_cos2 $integer $precision $delta]}
        3 {set l [_sin2 $integer $precision $delta]}
        default {error "internal error"}
    }
    # precision -> exp (multiplied by -1)
    lset l 1 [expr {-([lindex $l 1])}]
    lset l 0 0 $signe
    return [normalize [linsert $l 0 F]]
}

################################################################################
# compute cos(x) where 0<=x<Pi/2
# returns : a list formed with :
# 1. the mantissa
# 2. the precision (opposite of the exponent)
# 3. the uncertainty (doubt range)
################################################################################
proc ::math::bigfloat::_cos2 {x precision delta} {
    set pi [_pi $precision]
    set pis4 [rnddiv $pi {1 4}]
    set pis2 [rnddiv $pi {1 2}]
    if {[intCompare $x $pis4]>=0} {
        # cos(Pi/2-x)=sin(x)
        set x [intSub $pis2 $x]
        set delta [_intAdd $delta 1]
        return [_sin $x $precision $delta]
    }
    return [_cos $x $precision $delta]
}

################################################################################
# compute cos(x) where 0<=x<Pi/4
# returns : a list formed with :
# 1. the mantissa
# 2. the precision (opposite of the exponent)
# 3. the uncertainty (doubt range)
################################################################################
proc ::math::bigfloat::_cos {x precision delta} {
    set un [lshift {1 1} $precision]
    # Taylor development follows :
    # cos(x)=1-x^2/2 + x^4/4! ... + (-1)^(2n)*x^(2n)/2n!
    set s $un
    # 'd' is the uncertainty on x^2
    set d [intMul $x [intMul $delta 2]]
    set d [rshift $d $precision]
    # x=x^2 (because in the Taylor development, there are only even powers of x)
    set x [rshift [intMul $x $x] $precision]
    set denom1 1
    set denom2 2
    set signe -1
    set t [_intDiv $x 2]
    set delta {1 0}
    set dt $d
    while {![intIsZero $t]} {
        if {$signe<0} {
            set s [intSub $s $t]
        } else  {
            set s [intAdd $s $t]
        }
        set delta [intAdd $delta $dt]
        incr denom1 2
        incr denom2 2
        set dt [rshift [intAdd [intMul $x $dt]\
                [intMul $t $d]] $precision]
        set t [rshift [intMul $x $t] $precision]
        set t [_intDiv $t [expr {$denom1*$denom2}]]
        set signe [expr {-$signe}]
    }
    return [list $s $precision $delta]
}

################################################################################
# cotangent : the trivial algorithm is used
################################################################################
proc ::math::bigfloat::cotan {x} {
    return [::math::bigfloat::div [::math::bigfloat::cos $x] [::math::bigfloat::sin $x]]
}

################################################################################
# converts angles from degrees to radians
################################################################################
proc ::math::bigfloat::deg2rad {x} {
    checkFloat x
    set xLen [expr {-[lindex $x 2]}]
    if {$xLen<1} {
        error "number too loose to convert to radians"
    }
    set pi [pi $xLen]
    return [div [mul $x $pi] [fromstr 180]]
}



################################################################################
# private proc to get : x modulo Pi/2
# and the quotient (x divided by Pi/2)
# used by cos , sin & others
################################################################################
proc ::math::bigfloat::divPiQuarter {integer precision} {
    set dpi [rshift [_intMul [_pi [expr {1+$precision}]] 2]]
    # modulo 2Pi
    foreach {n integer} [intDivqr $integer $dpi] {break}
    # fin modulo 2Pi
    set pi [_pi $precision]
    foreach {n integer} [intDivqr $integer $pi] {break}
    # now divide by Pi/2
    set n [intMul $n {1 2}]
    set pis2 [rnddiv $pi {1 2}]
    foreach {m integer} [intDivqr $integer $pis2] {break}
    return [list [intAdd $n $m] $integer]
}


################################################################################
# divide A by B and returns the result
# throw error : divide by zero
################################################################################
proc ::math::bigfloat::div {a b} {
    checkNumber a b
    # dispatch to an appropriate procedure 
    if {[isInt $a]} {
        if {[isInt $b]} {
            return [intDiv $a $b]
        }
        error "trying to divide a BigInt by a BigFloat"
    }
    if {[isInt $b]} {return [divFloatByInt $a $b]}
    foreach {dummy integerA expA deltaA} $a {break}
    foreach {dummy integerB expB deltaB} $b {break}
    # computes the limits of the doubt (or uncertainty) interval
    set BMin [intSub $integerB $deltaB]
    set BMax [intAdd $integerB $deltaB]
    if {[intSign $integerB]<0} {
        # swap BMin and BMax
        set temp $BMin
        set BMin $BMax
        set BMax $temp
    }
    # multiply by zero gives zero
    if {[intIsZero $integerA]} {
        # why not return any number or the integer 0 ?
        # because there is an exponent that might be different between two BigFloats
        # 0.00 --> exp = -2, 0.000000 -> exp = -6
        return $a 
    }
    # test of the division by zero
    if {[intSign $BMin]*[intSign $BMax]<0 || \
                [intIsZero $BMin] || [intIsZero $BMax]} {
        error "divide by zero"
    }
    # shift A because we need accuracy
    # example : A=100e-2 and B=230e-4
    # if we divide 100 by 230 we get 0 as a result
    # but if we shift A's mantissa by 3 digits we get ..
    # 100000/230= 434
    # then exponent is computed as : (-2)-3-(-4)=-1
    # the number we get is 434e-1=43.4
    set l [length $integerB]
    set integerA [lshift $integerA $l]
    set deltaA [lshift $deltaA $l]
    set exp [expr {$expA-$l-$expB}]
    # relative uncertainties (dX/X) are added
    # to give the relative uncertainty of the result
    # i.e. 3% on A + 2% on B --> 5% on the quotient
    # d(A/B)/(A/B)=dA/A + dB/B
    # Q=A/B
    # dQ=dA/B + dB*A/B*B
    # dQ is "delta"
    set delta [incrdiv [intMul $deltaB [abs $integerA]] [abs $integerB]]
    set delta [incrdiv [intAdd $delta $deltaA] [abs $integerB]]
    set quotient [intDiv $integerA $integerB]
    if {[intSign $integerB]*[intSign $integerA]<0} {
        set quotient [_intAdd $quotient -1]
    }
    return [normalize [list F $quotient $exp $delta]]
}




################################################################################
# divide a BigFloat A by a BigInt B
# throw error : divide by zero
################################################################################
proc ::math::bigfloat::divFloatByInt {a b} {
    # type check
    checkFloat a
    if {![isInt $b]} {
        error "'$b' is not a BigInt"
    }
    foreach {dummy integer exp delta} $a {break}
    # zero divider test
    if {[intIsZero $b]} {
        error "divide by zero"
    }
    # shift left for accuracy ; see other comments in [div] procedure
    set l [length $b]
    set integer [lshift $integer $l]
    set delta [lshift $delta $l]
    incr exp -$l
    set integer [intDiv $integer $b]
    # the uncertainty is always evaluated to the ceil value
    # and as an absolute value
    set delta [incrdiv $delta [abs $b]]
    return [normalize [list F $integer $exp $delta]]
}





################################################################################
# returns 1 if A and B are equal, 0 otherwise
# IN : a, b (BigFloats)
################################################################################
proc ::math::bigfloat::equal {a b} {
    if {[isInt $a] && [isInt $b]} {
        return [expr {[intCompare $a $b]==0}]
    }
    checkFloat a b
    foreach {dummy aint aexp adelta} $a {break}
    foreach {dummy bint bexp bdelta} $b {break}
    # set all BigInt to the same level (exponent)
    # with lshift
    set diff [expr {$aexp-$bexp}]
    if {$diff<0} {
        set diff [expr {-$diff}]
        set bint [lshift $bint $diff]
        set bdelta [lshift $bdelta $diff]
    } elseif {$diff>0} {
        set aint [lshift $aint $diff]
        set adelta [lshift $adelta $diff]
    }
    # compute limits of the number's doubt range
    set asupInt [intAdd $aint $adelta]
    set ainfInt [intSub $aint $adelta]
    set bsupInt [intAdd $bint $bdelta]
    set binfInt [intSub $bint $bdelta]
    # A & B are equal
    # if their doubt ranges overlap themselves
    if {[intCompare $bint $aint]==0} {
        return 1
    }
    if {[intCompare $bint $aint]>0} {
        set r [expr {[intCompare $asupInt $binfInt]>=0}]
    } else {
        set r [expr {[intCompare $bsupInt $ainfInt]>=0}]
    }
    return $r
}

################################################################################
# returns exp(X) where X is a BigFloat
################################################################################
proc ::math::bigfloat::exp {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    if {$exp>=0} {
        incr exp
        set integer [lshift $integer $exp]
        set delta [lshift $delta $exp]
        set exp -1
    }
    set precision [expr {-$exp}]
    # add 1 digit of precision for safety
    incr precision
    set integer [lshift $integer]
    set delta [lshift $delta]
    set log10 [_log10 $precision]
    foreach {new_exp integer} [intDivqr $integer $log10] {break}
    # new_exp = integer part of x/log(10)
    # integer = remainder
    # exp(K.log(10)+r)=10^K.exp(r)
    # so we just have to compute exp(r), r is small so
    # the Taylor development will converge quickly
    set delta [intAdd $delta $new_exp]
    foreach {integer delta} [_exp $integer $precision $delta] {break}
    set delta [rshift $delta]
    incr precision -1
    # multiply by 10^K , and take care of the sign
    # example : X=-6.log(10)+0.01
    # exp(X)=exp(0.01)*1e-6
    if {[llength $new_exp]>2} {error "floating-point overflow due to exp"}
    set new_exp [tostr $new_exp]
    set exp [expr {$new_exp-$precision}]
    return [normalize [list F [rshift $integer] $exp $delta]]
}


################################################################################
# private procedure to compute exponentials
# using Taylor development of exp(x) :
# exp(x)=1+ x + x^2/2 + x^3/3! +...+x^n/n!
# input : integer (the mantissa)
#         precision (the number of decimals)
#         delta (the doubt limit, or uncertainty)
# returns a list : 1. the mantissa of the result
#                  2. the doubt limit, or uncertainty
################################################################################
proc ::math::bigfloat::_exp {integer precision delta} {
    if {[intIsZero $integer]} {
        return [list [lshift {1 1} $precision] $delta]
    }
    set one [lshift {1 1} $precision]
    set s [intAdd $one $integer]
    set d [_intAdd [_intDiv $delta 2] 1]
    set delta [intAdd $delta $delta]
    # dt = uncertainty on x^2
    set dt [rshift [intMul $d $integer] $precision]
    # t= x^2/2
    set t [rshift [intMul $integer $integer] $precision]
    set t [_intDiv $t 2]
    set denom 2
    while {![intIsZero $t]} {
        # the sum is called 's'
        set s [intAdd $s $t]
        set delta [intAdd $delta $dt]
        # we do not have to keep trace of the factorial, we just iterate divisions
        incr denom
        # add delta
        if {[llength $d]!=2 || [lindex $d 1]==1} {
            set d [_intAdd [_intDiv $d $denom] 1]
        }
        set dt [intAdd $dt $d]
        set t [rshift [intMul $integer $t] $precision]
        # here we divide
        set t [_intDiv $t $denom]
    }
    return [list $s $delta]
}


################################################################################
# procedure floor : identical to [expr floor($x)] in functionality
# arguments : number IN (a BigFloat)
# returns : the floor value as a BigInt
################################################################################
proc ::math::bigfloat::floor {number} {
    checkFloat number
    set number [normalize $number]
    if {[isnull $number]} {
        # returns the BigInt 0
        return {1 0}
    }
    foreach {dummy integer exp delta} $number {break}
    if {$exp>0} {
        error "not enough precision to perform rounding (floor)"
    }
    set decimal [lshift {1 1} [expr {-$exp}]]
    foreach {integer last} [intDivqr $integer $decimal] {break}
    if {[intSign $integer]<0} {
        set last [intSub $decimal $last]
    }
    # if sign>0 : nothing to do (return integer part)
    # if sign<0 :
    #                - if last=0 nothing to do
    #                - if last<>0 decrement integer part
    if {[intSign $integer]<0 && [intCompare [abs $last] $delta]<0} {
        set integer [_intAdd $integer 1]
    }
    return $integer
}


################################################################################
# returns a list formed by an integer and an exponent
# x = (A +/- C) * 10 power B
# return [list "F" A B C] (where F is the BigFloat tag)
# A and C are BigInts, B is a raw integer
# return also a BigInt when there is neither a dot, nor a 'e' exponent
# structure of a BigInt : [list sign last lastbutone .. first]
# sign : the sign of the integer ; other numbers are chunks of 9 digits
################################################################################
proc ::math::bigfloat::fromstr {string {trailingZeros 0}} {
    set signe 1
    if {$trailingZeros<0} {
        error "second argument has to be a positive integer"
    }
    # eliminate the sign problem
    if {[string index $string 0]=="-"} {
        set signe -1
        set string [string range $string 1 end]
    }
    # integer case (not a floating-point number)
    if {[string is digit $string]} {
        set liste [list $signe]
        set s [expr {$::math::bigfloat::segmentSize-1}]
        set len [string length $string]
        for {set i 0} {($i+1)*$::math::bigfloat::segmentSize<$len} {incr i} {
            lappend liste [string range $string end-$s end]
            set string [string range $string 0 \
                    end-$::math::bigfloat::segmentSize]
        }
        lappend liste $string
        if {$trailingZeros!=0} {
            set liste [lshift $liste $trailingZeros]
        }
        return [Dec $liste]
    }
    # floating-point number : check for an exponent
    # scientific notation
    set tab [split $string e]
    if {[llength $tab]>2} {error "syntax error in number : $string"}
    if {[llength $tab]==2} {
        set exp [lindex $tab 1]
        set number [lindex $tab 0]
    } else {
        set exp 0
        set number [lindex $tab 0]
    }
    # a floating-point number may have a dot
    set tab [split $number .]
    if {[llength $tab]>2} {error "syntax error in number : $string"}
    if {[llength $tab]==2} {
        set number [lindex $tab 0]
        set fin [lindex $tab 1]
        incr exp -[string length $fin]
    }
    append number $fin
    # this is necessary to ensure we can call fromstr (recursively) with
    # the mantissa ($number)
    if {![string is digit $number]} {
        error "$number is not a number"
    }
    incr exp -$trailingZeros
    set number [lshift [fromstr $number] $trailingZeros]
    lset number 0 $signe
    # the F tags a BigFloat
    # a BigInt in internal representation begins by the sign
    # delta is 1 as a BigInt
    return [list F $number $exp {1 1}]
}

################################################################################
# returns 1 if x is a BigFloat, 0 elsewhere
################################################################################
proc ::math::bigfloat::isFloat {x} {
    # a BigFloat is a list of : "F" mantissa exponent delta
    if {[llength $x]!=4} {
        return 0 
    }
    # the marker is the letter "F"
    if {[string equal [lindex $x 0] F]} {
        return 1
    }
    return 0
}



################################################################################
# returns 1 if x is null, 0 otherwise
################################################################################
proc ::math::bigfloat::isnull {x} {
    if {[isInt $x]} {
        return [equal $x {1 0}]
    }
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    lset integer 0 1
    if {[intCompare $delta $integer]>=0} {return 1}
    return 0
}


################################################################################
# compute log(X)
################################################################################
proc ::math::bigfloat::log {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    set integer [Dec $integer]
    if {[intCompare $integer {1 0}]<=0} {
        error "zero logarithm error"
    }
    if {[isnull $x]} {
        error "number is null"
    }
    set precision [length $integer]
    set five [lshift {1 5} [expr {$precision-1}]]
    set delta [_intAdd [_logOnePlusEpsilon $delta $integer $precision] 1]
    set double 0
    # now : x belongs to the interval [0.1,1.0[
    # we double x until it belongs to the interval [0.5,1.0[
    while {[intCompare $integer $five]<0} {
        incr double
        set integer [_intMul $integer 2]
        # the next line is not needed (11/01/2005)
        #set delta [_intMul $delta 2]
    }
    # now 0.5<= x/10^precision <1.0
    # attention : _log returns a positive integer
    foreach {integer deltaIncr} [_log $integer] {break}
    set delta [intAdd $delta $deltaIncr]
    # log(a * 2^-double)= log(a) - double*log(2)
    # result = log(x) - double*log(2)
    set integer [opp [intAdd [_intMul [_log2 $precision] $double] $integer]]
    set delta [_intAdd $delta $double]
    incr exp $precision
    if {$exp!=0} {
        set integer [intAdd [_intMul [_log10 $precision] $exp] $integer]
        set delta [_intAdd $delta [expr {abs($exp)}]]
    }
    return [normalize [list F $integer [expr {-$precision}] $delta]]
}

proc ::math::bigfloat::_logOnePlusEpsilon {epsNum epsDenom precision} {
    if {[intCompare $epsNum $epsDenom]>=0} {
        error "number is null"
    }
    set s [lshift $epsNum $precision]
    set s [intDiv $s $epsDenom]
    set divider 2
    set t [intDiv [intMul $s $epsNum] $epsDenom]
    set u [_intDiv $t $divider]
    while {![intIsZero $u]} {
        set s [_intAdd [intAdd $s $u] 1]
        incr divider
        set t [intDiv [intMul $t $epsNum] $epsDenom]
        set u [_intDiv $t $divider]
    }
    return $s
}


# compute log(0.integer)
proc ::math::bigfloat::_log {integer} {
    set integer [lshift $integer 3]
    set delta [list 1 3]
    set l [length $integer]
    set un [lshift [list 1 1] $l]
    set integer [intSub $un $integer]
    set s $integer
    set t [intMul $integer $integer]
    set t [rshift $t $l]
    set denom 2
    set u [_intDiv $t $denom]
    while {![intIsZero $u]} {
        set s [intAdd $s $u]
        set delta [_intAdd $delta 1]
        set t [rshift [intMul $t $integer] $l]
        incr denom
        set u [_intDiv $t $denom]
    }
    return [list [roundshift $s 3] [_intAdd [rshift $delta 3] 1]]
}

proc ::math::bigfloat::__log {num denom precision} {
    # p is the precision
    # pk is the precision increment
    # 10 power pk is also the maximum number of iterations
    # for a number close to 1 but lower than 1,
    # (denom-num)/denum is (in our case) lower than 1/5
    # so the maximum nb of iterations is for:
    # 1/5*(1+1/5*(1/2+1/5*(1/3+1/5*(...))))
    # the last term is 1/n*(1/5)^n
    # for the last term to be lower than 10^(-p-pk)
    # the number of iterations has to be
    # 10^(-pk).(1/5)^(10^pk) < 10^(-p-pk)
    # log(1/5).10^pk < -p
    # 10^pk > p/log(5)
    # pk > log(10)*log(p/log(5))
    # now set the variable n to the precision increment i.e. pk
    set n [expr {int(log(10)*log($precision/log(5)))+1}]
    incr precision $n
    set num [list 1 [expr {$denom-$num}]]
    set denom [list 1 $denom]
    set s [intDiv [lshift $num $precision] $denom]
    set t [intDiv [intMul $s $num] $denom]
    set d 2
    set u [_intDiv $t $d]
    while {![intIsZero $u]} {
        set s [intAdd $s $u]
        set t [intDiv [intMul $t $num] $denom]
        incr d
        set u [_intDiv $t $d]
    }
    return [rshift $s $n]
}


# compute log(10)
proc ::math::bigfloat::_log10 {precision} {
    incr precision 2
    if {[catch {set a $::math::bigfloat::log10}]} {
        __logbis $precision
    } else {
        set l [length $::math::bigfloat::log10]
        if {$precision>$l-1} {
            __logbis $precision
        }
    }
    incr precision -1
    return [_round $::math::bigfloat::log10 $precision]
}

proc ::math::bigfloat::__logbis {precision} {
    incr precision 2
    # ln(2)=3*ln(1-4/5)+ln(1-125/128)
    set a [__log 125 128 $precision]
    set b [__log 4 5 $precision]
    set r [intAdd [_intMul $b 3] $a]
    set ::math::bigfloat::log2 [rshift $r 2]
    # ln(10)=10.ln(1-4/5)+3*ln(1-125/128)
    set r [intAdd [lshift $b] [_intMul $a 3]]
    set ::math::bigfloat::log10 [rshift $r 2]
}


# compute log(2)
proc ::math::bigfloat::_log2 {precision} {
    if {![info exists ::math::bigfloat::log2]} {
        __logbis $precision
    } else {
        set l [length $::math::bigfloat::log2]
        if {$precision>$l} {
            __logbis $precision
        }
    }
    return [_round $::math::bigfloat::log2 $precision]
}


################################################################################
# returns A modulo B (like with fmod() math function
################################################################################
proc ::math::bigfloat::mod {a b} {
    checkNumber a b
    if {[isInt $a] && [isInt $b]} {return [intMod $a $b]}
    if {[isInt $a]} {error "trying to divide a BigInt by a BigFloat"}
    set quotient [div $a $b]
    if {[intSign [floor $quotient]]>0} {
        set quotient [floor $quotient]
    } else {
        set quotient [ceil $quotient]
    }
    return [sub $a [mul $quotient $b]]
}

################################################################################
# returns A times B
################################################################################
proc ::math::bigfloat::mul {a b} {
    checkNumber a b
    if {[isInt $a]} {
        if {[isInt $b]} {
            return [intMul $a $b]
        }
        return [mulFloatByInt $b $a]
    }
    if {[isInt $b]} {return [mulFloatByInt $a $b]}
    foreach {dummy integerA expA deltaA} $a {break}
    foreach {dummy integerB expB deltaB} $b {break}
    set exp [expr {$expA+$expB}]
    set integer [intMul $integerA $integerB]
    set deltaA [intMul [abs $integerB] $deltaA]
    set deltaB [intMul [abs $integerA] $deltaB]
    set delta [intAdd $deltaA $deltaB]
    #puts $delta
    return [normalize [list F $integer $exp $delta]]
}

################################################################################
# returns A times B, where B is a positive integer
################################################################################
proc ::math::bigfloat::mulFloatByInt {a b} {
    checkFloat a
    foreach {dummy integer exp delta} $a {break}
    if {![isInt $b]} {
        error "second argument expected to be a BigInt"
    }
    set integer [intMul $integer $b]
    set delta [intMul $delta $b]
    return [normalize [list F $integer $exp $delta]]
}

################################################################################
# normalizes a number : delta (incertitude)
# has to be one digit only to avoid increases of
# the memory footprint of the number
################################################################################
proc ::math::bigfloat::normalize {number} {
    if {[isInt $number]} {
        return [Dec $number]
    }
    foreach {dummy integer exp delta} $number {break}
    set integer [Dec $integer]
    set delta [Dec $delta]
    set l [length $delta]
    if {$l>2} {
        incr l -2
        set delta [_intAdd [rshift $delta $l] 1]
        set integer [rshift $integer $l]
        incr exp $l
    }
    return [list F $integer $exp $delta]
}



################################################################################
# returns -A (the opposite)
################################################################################
proc ::math::bigfloat::opp {a} {
    checkNumber a
    if {[isnull $a]} {
        return $a
    }
    if {[isInt $a]} {
        lset a 0 [expr {-[lindex $a 0]}]
        return $a
    }
    lset a 1 0 [expr {-[lindex $a 1 0]}]
    return $a
}

################################################################################
# gets Pi with precision digits
# after the dot (after you call [tostr] on the result)
################################################################################
proc ::math::bigfloat::pi {precision} {
    if {[llength $precision]>1} {
        if {[isInt $precision]} {
            set precision [tostr $precision]
        } else {
            error "'[tostr $precision]' expected to be an integer"
        }
        
    }
    return [list F [_pi $precision] -$precision [list 1 1]]
}

proc ::math::bigfloat::_pi {precision} {
    incr precision
    if {![info exists ::math::bigfloat::_pi0]} {
        set ::math::bigfloat::_pi0 [__pi $precision]
    } elseif {[length $::math::bigfloat::_pi0]<$precision} {
        set ::math::bigfloat::_pi0 [__pi $precision]
    }
    if {[length $::math::bigfloat::_pi0]==$precision} {
        return $::math::bigfloat::_pi0
    }
    return [_round $::math::bigfloat::_pi0 $precision]
}

proc ::math::bigfloat::__pi {precision} {
    incr precision 3
    set a [_intMul [_atanfract 18 $precision] 48]
    set a [intAdd $a [_intMul [_atanfract 57 $precision] 32]]
    set a [intSub $a [_intMul [_atanfract 239 $precision] 20]]
    incr precision -3
    return [_round $a $precision]
}

proc ::math::bigfloat::_round {integer precision} {
    return [rshift $integer [expr {[length $integer] - $precision}]]
}

################################################################################
# returns A power B, where B is a positive integer
################################################################################
proc ::math::bigfloat::pow {a b} {
    checkNumber a
    if {![isInt $b]} {
        error "pow : exponent is not a positive integer"
    }
    set res [list 1 1]
    while {1} {
        foreach {b reste} [intDivqr $b [list 1 2]] {break}
        if {![intIsZero $reste]} {
            set res [mul $res $a]
        }
        if {[intIsZero $b]} {
            return [normalize $res]
        }
        set a [mul $a $a]
    }
}

################################################################################
# converts angles for radians to degrees
################################################################################
proc ::math::bigfloat::rad2deg {x} {
    checkFloat x
    set xLen [expr {-[lindex $x 2]}]
    if {$xLen<1} {
        error "number too loose to convert to degrees"
    }
    set pi [pi $xLen]
    return [div [mul $x [fromstr 180]] $pi]
}

################################################################################
# retourne la partie entire (ou 0) du nombre "number"
################################################################################
proc ::math::bigfloat::round {number} {
    checkFloat number
    set number [normalize $number]
    # fetching integers (or BigInts) from the internal representation
    foreach {dummy integer exp delta} $number {break}
    if {[isnull $number]} {
        return [list 1 0]
    }
    if {$exp>0} {
        error "not enough precision to round (in round)"
    }
    set integer [roundshift $integer [expr {-$exp}]]
    return $integer
}



################################################################################
# gets sin(x)
################################################################################
proc ::math::bigfloat::sin {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    if {$exp>-2} {
        error "sin : not enough precision"
    }
    set precision [expr {-$exp}]
    # sin(2kPi+x)=sin(x)
    foreach {n integer} [divPiQuarter $integer $precision] {break}
    set delta [intAdd $delta $n]
    set d [intMod $n [list 1 4]]
    # maintenant integer>=0
    # sin(2Pi-x)=-sin(x)
    # sin(Pi-x)=sin(x)
    # sin(Pi/2+x)=cos(x)
    set signe 1
    switch  -- [tostr $d] {
        0 {set l [_sin2 $integer $precision $delta]}
        1 {set l [_cos2 $integer $precision $delta]}
        2 {set signe -1;set l [_sin2 $integer $precision $delta]}
        3 {set signe -1;set l [_cos2 $integer $precision $delta]}
        default {error "internal error"}
    }
    # precision --> exponent (the opposite)
    lset l 1 [expr {-([lindex $l 1])}]
    lset l 0 0 $signe
    return [normalize [concat F $l]]
}

proc ::math::bigfloat::_sin2 {x precision delta} {
    set pi [_pi [expr {1+$precision}]]
    set pis2 [rshift [_intDiv $pi 2]]
    set pis4 [rshift [_intDiv $pi 4]]
    if {[intCompare $x $pis4]>=0} {
        # sin(Pi/2-x)=cos(x)
        set delta [_intAdd $delta 1]
        set x [intSub $pis2 $x]
        return [_cos $x $precision $delta]
    }
    return [_sin $x $precision $delta]
}

# sin(x) avec x positif inferieur a Pi/4
proc ::math::bigfloat::_sin {x precision delta} {
    set s $x
    set double [rshift [_intMul [intMul $x $delta] 2] $precision]
    set x [rshift [intMul $x $x] $precision]
    set dt [rshift [intAdd [intMul $x $delta] [intMul $s $double]] $precision]
    set t [rshift [intMul $s $x] $precision]
    set denom2 2
    set denom3 3
    set t [_intDiv $t 6]
    set signe -1
    while {![intIsZero $t]} {
        if {$signe<0} {
            set s [intSub $s $t]
        } else  {
            set s [intAdd $s $t]
        }
        set delta [intAdd $delta $dt]
        set signe [expr {-$signe}]
        incr denom2 2
        incr denom3 2
        set dt [rshift [intAdd [intMul $x $dt] [intMul $t $double]] $precision]
        set t [rshift [intMul $t $x] $precision]
        set dt [intAdd $dt $double]
        set t [_intDiv $t [expr {$denom2*$denom3}]]
    }
    return [list $s $precision $delta]
}

################################################################################
# square root with Newton-Raphson algorithm
################################################################################
proc ::math::bigfloat::OldSqrt {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    # si x=0, retourner 0
    if {[isnull $x]} {
        return [list F [list 1 0] $exp [list 1 1]]
    }
    # we cannot get sqrt(x) if x<0
    if {[lindex $integer 0]<0} {
        error "negative sqrt : $x"
    }
    # Newton-Raphson method :
    # X_1=(X_0+1)/2 and X_N+1=(X_N+X_0/X_N)/2
    set x_initial $x
    set x_suivant [div [add $x [list 1 1]] [list 1 2]]
    while {![equal $x $x_suivant]} {
        set x $x_suivant
        set x_suivant [div [add $x [div $x_initial $x]] [list 1 2]]
    }
    return $x
}

################################################################################
# experimental procedure for extracting the square root of a BigFloat
################################################################################
proc ::math::bigfloat::sqrt {x} {
    checkFloat x
    foreach {dummy integer exp delta} $x {break}
    # si x=0, retourner 0
    if {[isnull $x]} {
        return [list F [list 1 0] $exp [list 1 1]]
    }
    # we cannot get sqrt(x) if x<0
    if {[lindex $integer 0]<0} {
        error "negative sqrt : $x"
    }
    # (n +/- delta)^1/2=n^1/2 * (1 +/- delta/n)^1/2
    # delta_final=n^1/2 * Sum(i from 1 to infinity)(delta/n)^i*(3*5*...*(2i-3))/(i!*2^i)
    foreach {delta precision} [_sqrtOnePlusEpsilon $delta $integer] {break}
    set intLen [length $integer]
    # intLen + exp = number of digits before the dot
    set precision [expr {$precision-($intLen+$exp)}]
    # square root extraction (like by hand)
    set integer [lshift $integer $intLen]
    incr exp -$intLen
    incr intLen $intLen
    if {$exp&1} {
        incr exp -1
        set integer [lshift $integer]
        incr intLen
    }
    if {!($intLen&1)} {
        # intLen is odd
        set firstSegment 2
    } else {
        # intLen is even
        set firstSegment 1
    }
    set numerator [rshift $integer [expr {$intLen-$firstSegment}]]
    set root [fromstr [expr {int(sqrt([tostr $numerator]))}]]
    set integer [Dec [intSub $integer [lshift [intMul $root $root] [expr {$intLen-$firstSegment}]]]]
    # now (denom*10+X)*X<=numerator
    # where X is just one digit
    set denominator [_intMul [lshift $root] 2]
    set rootLen 2
    incr intLen -$firstSegment
    while {$intLen>=$rootLen} {
        set numerator [rshift $integer [expr {$intLen-$rootLen}]]
        foreach {quotient remainder} [intDivqr $numerator $denominator] {break}
        # denminatorX*X<=numerator
        # quotient = numerator/denominator = X_temp
        # (X_temp*(denominator)+ X_temp^2) shall be lower than numerator
        # numerator - remainder + quotient^2 <=numerator
        # quotient^2 <= remainder
        if {[intCompare $remainder [intMul $quotient $quotient]]>0} {
            set quotient [_intAdd $quotient -1]
        }
        set root [intAdd [lshift $root] $quotient]
        set denominator [intAdd $denominator $quotient]
        set integer [Dec [intSub $integer [lshift [intMul $denominator $quotient] [expr {$intLen-$rootLen}]]]]
        set denominator [_intMul [lshift $root] 2]
        incr rootLen
        incr intLen -1
    }
    # root has exp, delta has -(precision)
    # an example : root= 2000e-4 delta= 1e-2
    # delta*root=20e-4
    set delta [_intAdd [rshift [intMul $delta $root] $precision] 1]
    return [normalize [list F $root [expr {$exp/2}] $delta]]
}



################################################################################
# compute abs(sqrt(1-delta/integer)-1)
# the returned value is a relative uncertainty
################################################################################
proc ::math::bigfloat::_sqrtOnePlusEpsilon {delta integer} {
    set l [length $integer]
    set x [intDiv [lshift $delta $l] $integer]
    set two [list 1 2]
    set fact $two
    # eps/2
    set result [intDiv $x $two]
    # eps^2/(2!*2)
    set temp [intDiv [intDiv [intDiv [intMul $result $delta] $integer] $two] $fact]
    set result [intAdd $result $temp]
    set numerator [list 1 3]
    set fact [_intAdd $fact 1]
    # (eps^3)*3/(3!*2^2)
    set temp [intDiv [intDiv [intMul [intDiv [intMul $temp $delta] $integer] $numerator] $two] $fact]
    while {![intIsZero $temp]} {
        set result [intAdd $result $temp]
        set numerator [_intAdd $numerator 2]
        set fact [_intAdd $fact 1]
        # u_n+1= u_n*(2n+1)/2n*eps
        set temp [intDiv [intDiv [intMul [intDiv [intMul $temp $delta] $integer] $numerator] $two] $fact]
    }
    return [list $result $l]
}

################################################################################
# substracts B to A
################################################################################
proc ::math::bigfloat::sub {a b} {
    checkNumber a b
    if {[isInt $a]} {
        if {[isInt $b]} {
            return [intSub $a $b]
        }
        return [add $a [opp $b]]
    }
    if {[isInt $b]} {
        return [opp [add [opp $a] $b]]
    }
    return [add $a [opp $b]]
}

################################################################################
# tangent
################################################################################
proc ::math::bigfloat::tan {x} {
    return [::math::bigfloat::div [::math::bigfloat::sin $x] [::math::bigfloat::cos $x]]
}



################################################################################
# converts a BigInt to a double (basic floating-point type)
# with respect to the global tcl_precision
################################################################################
proc ::math::bigfloat::todouble {x} {
    global tcl_precision
    checkFloat x
    set x [normalize $x]
    foreach {dummy integer exp delta} $x {break}
    set integer [Dec $integer]
    set l [length $integer]
    set tronquer [expr {$l-$tcl_precision}]
    if {$tronquer<=0} {
        return [tostr $x]
    }
    set int [Dec [roundshift $integer $tronquer]]
    if {[length $int]>$tcl_precision} {
        incr tronquer
    }
    set dt [rshift $delta $tronquer]
    while {![intIsZero $dt]} {
        incr tronquer
        set dt [rshift $delta $tronquer]
    }
    incr exp $tronquer
    set integer [roundshift $integer $tronquer]
    set isZero [intIsZero $integer]
    set nombre ""
    set filter "%0.${::math::bigfloat::segmentSize}ld"
    foreach n [lrange $integer 1 end] {
        set nombre [format $filter $n]$nombre
    }
    set nombre [trimleft $nombre]
    set l [string length $nombre]
    if {$exp>0} {
        incr exp $l
        incr exp -1
        set nombre [string index $nombre 0].[string range $nombre 1 end]
        append nombre "e+$exp"
    } elseif {$exp==0} {
        append nombre .
    } else {
        set exp [expr {-$exp}]
        if {$exp < $l} {
            set n [string range $nombre 0 end-$exp]
            incr exp -1
            append n .[string range $nombre end-$exp end]
            set nombre $n
        } elseif {$l==$exp} {
            set nombre "0.$nombre"
        } else  {
            set nombre "[string index $nombre 0].[string range $nombre 1 end]e-[expr {$exp-$l+1}]"
        }
    }
    if {[lindex $integer 0]<0 && !$isZero} {set nombre "-$nombre"}
    return $nombre
}

################################################################################
# converts a number stored as a list to a string in which all digits are true
################################################################################
proc ::math::bigfloat::tostr {number} {
    if {[isInt $number]} {
        set filter "%0.${::math::bigfloat::segmentSize}ld"
        set strInt ""
        foreach n [lrange $number 1 end] {
            set strInt [format $filter $n]$strInt
        }
        set strInt [trimleft $strInt]
        if {[lindex $number 0]<0} {
            return -$strInt
        }
        return $strInt
    }
    checkFloat number
    foreach {dummy integer exp delta} $number {break}
    set integer [Dec $integer]
    set signe [lindex $integer 0]
    set result $integer
    # sign is saved, now let's use absolute value of number
    lset result 0 1
    # if {[intCompare $result 0]} {
    # if {[intCompare $result $delta]<=0} {
    # return 0
    # }
    # }
    # right shift until delta is null
    set decalage [length $delta]
    set result [roundshift $result $decalage]
    incr exp $decalage
    set isZero [intIsZero $result]
    set filter "%0.${::math::bigfloat::segmentSize}ld"
    set strInt ""
    foreach n [lrange $result 1 end] {
        set strInt [format $filter $n]$strInt
    }
    set result [trimleft $strInt]
    set l [string length $result]
    if {$exp>0} {
        incr exp $l
        incr exp -1
        set result [string index $result 0].[string range $result 1 end]
        append result "e+$exp"
    } elseif {$exp==0} {
        append result .
    } else {
        set exp [expr {-$exp}]
        if {$exp < $l} {
            set n [string range $result 0 end-$exp]
            incr exp -1
            append n .[string range $result end-$exp end]
            set result $n
        } elseif {$l==$exp} {
            set result "0.$result"
        } else  {
            set result "[string index $result 0].[string range $result 1 end]e-[expr {$exp-$l+1}]"
            
        }
    }
    if {$signe<0 && !$isZero} {set result "-$result"}
    return $result
}

################################################################################
# PART IV
# HYPERBOLIC FUNCTIONS
################################################################################

################################################################################
# hyperbolic cosinus
################################################################################
proc ::math::bigfloat::cosh {x} {
    return [div [add [exp $x] [exp [opp $x]]] [list 1 2]]
}

################################################################################
# hyperbolic sinus
################################################################################
proc ::math::bigfloat::sinh {x} {
    return [div [sub [exp $x] [exp [opp $x]]] [list 1 2]]
}

################################################################################
# hyperbolic tangent
################################################################################
proc ::math::bigfloat::tanh {x} {
    set up [exp $x]
    set down [exp [opp $x]]
    return [div [sub $up $down] [add $up $down]]
}

namespace eval ::math::bigfloat {
    foreach function {
        add mul sub div mod pow
        isnull compare equal
        fromstr tostr todouble isInt
        lshift rshift roundshift
        exp log sqrt round ceil floor
        sin cos tan cotan asin acos atan
        cosh sinh tanh abs opp
        pi deg2rad rad2deg
    } {
        namespace export $function
    }
}
namespace import ::math::bigfloat::*

package provide math::bigfloat 1.1
